;
; lg-dict.scm
;
; Link Grammar dictionary API
;
(define-module (opencog nlp lg-dict))

(use-modules (srfi srfi-1) (opencog) (opencog lg-config) (opencog nlp) (opencog exec))

(load-extension (string-append opencog-ext-path-lg-dict "liblg-dict") "opencog_nlp_lgdict_init")

; ---------------------------------------------------------------------

(export lg-conn-type-match?)
(set-procedure-property! lg-conn-type-match? 'documentation
"
  lg-conn-type-match? CON-A CON-B
     Return #t if connector CON-A matches CON-B, else return #f.
     Both CON-A and CON-B must be of type LgConnector.

     This does NOT check connector direction agreement; it only
     checks the connector strings, using the standard Link Grammar
     connector matching rules.
")

(export lg-conn-linkable?)
(set-procedure-property! lg-conn-linkable? 'documentation
"
  lg-conn-linkable? CON-A CON-B
     Return #t if connector CON-A can link to CON-B, else return #f.
     Both CON-A and CON-B must be of type LgConnector.

     This checks the connector strings for linkability, using the
     standard Link Grammar connector matching rules.
")

; ---------------------------------------------------------------------

(define-public (lg-dict-entry WORD)
"
  lg-dict-entry  WORD
     Fetch the dictionary entry for WORD and place it in the atomspace.
     WORD must be a WordNode. The langauge is assumed to be English.

     The dictionary entry can subsequently be obtained by calling
     (cog-incoming-by-type WORD 'LgDisjunct)
"
	(define djset (cog-incoming-by-type WORD 'LgDisjunct))
	(if (nil? djset)
		(let ((dentry (LgDictEntry WORD (LgDictNode "en"))))
			; XXX Argh. This function gets called in par-map in sureal,
			; and because it races, the dentry could already be deleted
			; by now, in another thread. If so, the the cog-extract! throws.
			; Yuck.
			(catch 'wrong-type-arg
				(lambda ()
					(cog-execute! dentry)
					(cog-extract! dentry))
				(lambda (key . args) #f))
			(cog-incoming-by-type WORD 'LgDisjunct)
		)
		djset)
)


(define-public (lg-get-dict-entry WORD)
"
  lg-get-dict-entry  WORD
     Fetch the dictionary entry for WORD and place it in the atomspace.
     WORD must be a WordNode.  The language is assumed to be English.

     The dictionary entry can subsequently be obtained by calling
     (cog-incoming-by-type WORD 'LgDisjunct)

     DEPRECATED! Use lg-dict-entry instead!
"
	(SetLink (lg-dict-entry WORD))
)

(define-public (lg-similar? word1 word2)
"
  lg-similar? WORD1 WORD2 - Check two words have a common disjunct.

  Caution: This utility isn't really useful, as designed. Most LG
  dictioary entries have hundreds, if not tens of thousands of
  disjuncts, and will frequently have at least one disjunct in common,
  while otherwise being dissimilar.  A much better measure of similarity
  is provided by the Jaccard distance, which counts how many disjuncts
  are shared in common.  (The Jaccard distance is currently not
  implemented here.)

  A second problem with this function is that the implementation is very
  inefficient and slow.
"
	(define (get-set w)
		(define roots (filter
			(lambda (l) (equal? (cog-type l) 'LgDisjunct))
			(cog-incoming-set w)))
		(map cog-get-partner roots (circular-list w))
	)

	; Create the dictionary entry as needed
	(lg-get-dict-entry word1)
	(lg-get-dict-entry word2)

	; Check if the two words have a common LG dict entry
	(not (nil? (lset-intersection equal? (get-set word1) (get-set word2))))
)

; ---------------------------------------------------------------------

(define-public (lg-conn-get-type conn)
"
  lg-conn-get-type CON - Get the LgConnNode out of LgConnector link
"
	(cog-outgoing-atom conn 0)
)

; ---------------------------------------------------------------------
(define-public (lg-conn-get-dir conn)
"
  lg-conn-get-dir CON - Get the LgConnDirNode out of LgConnector link
"
	(cog-outgoing-atom conn 1)
)

; ---------------------------------------------------------------------
